import json
from pathlib import Path
from typing import Any, Dict

from jsonschema import Draft7Validator
from jsonschema.exceptions import ValidationError
from yaml import safe_load

from pyinsights.exceptions import (
    ConfigInvalidSyntaxError,
    ConfigNotFoundError,
    ConfigVersionUnknownError,
    InvalidVersionError
)


ConfigType = Dict[str, Any]
SchemaType = Dict[str, Any]


def load_config(filepath: str) -> ConfigType:
    """Load the configuration file

    Arguments:
        filepath {str}

    Raises:
        ConfigNotFoundError

    Returns:
        config {ConfigType}
    """

    try:
        with open(filepath) as fd:
            config = safe_load(fd)

        return config
    except FileNotFoundError:
        raise ConfigNotFoundError('Could not find the configuration')


def load_schema(version: str) -> SchemaType:
    """Load the schema json file

    Arguments:
        version {str}

    Raises:
        InvalidVersionError

    Returns:
        schema {SchemaType}
    """

    basepath = Path(__file__).parent.resolve()
    filename = f'version_{version}.json'
    schema_filpath = f'{basepath}/schema/{filename}'

    try:
        with open(schema_filpath) as fd:
            schema = json.load(fd)

        return schema
    except FileNotFoundError:
        raise InvalidVersionError(f'The version {repr(version)} is invalid')


def validate(config: ConfigType) -> bool:
    """Validate the configuration

    Arguments:
        config {ConfigType}

    Raises:
        ConfigVersionUnknownError
        ConfigInvalidSyntaxError

    Returns:
        bool
    """

    try:
        version = config['version']
    except KeyError:
        raise ConfigVersionUnknownError('Please specify the version')

    schema = load_schema(version)

    try:
        Draft7Validator(schema).validate(config)
        config.pop('version')
    except ValidationError as err:
        raise ConfigInvalidSyntaxError(err)
    else:
        return True



import argparse
from typing import Any, Dict
import sys

from pyinsights import __version__
from pyinsights.pyinsights import run


def parse_args() -> Dict[str, Any]:
    """Parse arguments

    Returns:
        Dict[str, Any]
    """

    parser = argparse.ArgumentParser(
        prog='PyInsights',
        description='AWS CloudWatch Logs Insights is wrapped by Python',
    )

    parser.add_argument(
        '-c',
        '--config',
        required=True,
        default='pyinsights.yml',
        help='PyInsights config file path',
    )

    parser.add_argument('-p', '--profile', help='AWS profile name')

    parser.add_argument('-r', '--region', help='AWS region')

    parser.add_argument(
        '-v', '--version', action='version', version=__version__
    )

    return vars(parser.parse_args())


def cli() -> bool:
    args = parse_args()
    sys.exit(run(args))



import sys
import random
from datetime import datetime, timedelta
from typing import Dict, Type, Union

from pyinsights.exceptions import InvalidDurationError


DatetimeType = Type[datetime]


def convert_epoch(time: Union[str, DatetimeType]) -> int:
    """Convert datetime string to epoch (unix timestamp)

    Arguments:
        time {Union[str, DatetimeType]}
            -- datetime string that format is `%Y-%m-%d %H:%M:%S`

    Returns:
        epoch {int}
    """

    if isinstance(time, str):
        time_format = '%Y-%m-%d %H:%M:%S'
        try:
            time = datetime.strptime(time, time_format)
        except ValueError:
            raise InvalidDurationError(
                f'{repr(time)} is invalid as duration parameter'
            )

    epoch = int(time.timestamp())
    return epoch


time_unit_map = {
    's': 'seconds',
    'm': 'minutes',
    'h': 'hours'
}


def get_times(duration: str) -> Dict[str, DatetimeType]:
    try:
        unit = duration[-1]
        duration = int(duration[:-1])
        arg = {time_unit_map[unit]: duration}
    except (ValueError, IndexError, KeyError):
        raise InvalidDurationError(
            f'{repr(duration)} is invalid as duration parameter'
        )

    end_time = datetime.now()
    start_time = end_time - timedelta(**arg)
    times = {'start_time': start_time, 'end_time': end_time}
    return times


def color() -> str:
    """Choice a color

    Returns:
        str
    """

    colors = [
        Color.Red,
        Color.Green,
        Color.Yellow,
        Color.Blue,
        Color.Purple,
        Color.Cyan
    ]
    color = random.choice(colors)
    return color


class Color:
    Red = '\033[31m'
    Green = '\033[32m'
    Yellow = '\033[33m'
    Blue = '\033[34m'
    Purple = '\033[35m'
    Cyan = '\033[36m'


class Accessory:
    End = '\033[0m'
    Accent = '\033[01m'


def processing(msg: str, end: str = '') -> None:
    """Display processing on terminal

    Arguments:
        msg {str}

    Keyword Arguments:
        end {str} -- (default: {''})
    """

    processing_msg = f'{Accessory.Accent}{color()}{msg}{Accessory.End}{end}'
    sys.stdout.write(processing_msg)
    sys.stdout.flush()



import os
from typing import Any, Dict, List, Optional, Union

import boto3
import botocore.errorfactory

from pyinsights.exceptions import (
    QueryNotYetStartError,
    NotFetchQueryResultError,
)
from pyinsights.helper import convert_epoch, DatetimeType


AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY')
AWS_DEFAULT_REGION = os.getenv('AWS_DEFAULT_REGION')


class InsightsClient:
    def __init__(
        self,
        region: Optional[str] = None,
        profile: Optional[str] = None,
    ) -> None:
        """
        Keyword Arguments:
            region {Optional[str]}
            profile {Optional[str]}
        """

        region_name = region or AWS_DEFAULT_REGION

        session = boto3.Session(
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
            region_name=region_name,
            profile_name=profile,
        )

        self._client = session.client('logs')
        self._query_id = None

    def start_query(
        self,
        query_string: str,
        start_time: Union[str, DatetimeType],
        end_time: Union[str, DatetimeType],
        log_group_name: List[str],
        limit: Optional[int] = 1000,
    ) -> bool:
        """Start query

        Arguments:
            query_string {str} -- query string
            start_time {Union[str, DatetimeType]}
                -- datetime that format is `%Y-%m-%d %H:%M:%S`
            end_time {Union[str, DatetimeType]}
                -- datetime that format is `%Y-%m-%d %H:%M:%S`
            log_group_name {List[str]} -- aws cloudwatch log group to query

        Keyword Arguments:
            limit {Optional[int] -- the maximum number of log events
                                    to return in the query. (default: {1000})

        Returns:
            bool
        """

        start_time_epoch = convert_epoch(start_time)
        end_time_epoch = convert_epoch(end_time)

        response = self._client.start_query(
            logGroupNames=log_group_name,
            startTime=start_time_epoch,
            endTime=end_time_epoch,
            queryString=query_string,
            limit=limit,
        )

        self._query_id = response['queryId']
        return True

    def fetch_result(self) -> Dict[str, Any]:
        """Fetch the query result

        Raises:
            QueryNotYetStartError

        Returns:
            results {Dict[str, Any]}
        """

        if self._query_id is None:
            raise QueryNotYetStartError('The Query has not yet started')

        results = self._client.get_query_results(queryId=self._query_id)
        status = results['status']

        if status in ['Scheduled', 'Running']:
            results.update(self.fetch_result())

        elif status == 'Failed':
            raise NotFetchQueryResultError('Could not fetch the query result')

        return results

    def end_query(self) -> bool:
        """Gracefully terminate query

        Raises:
            err:
                raise exception if InvalidParameterException,
                ServiceUnavailableException or other error occurs

        Returns:
            bool
        """

        if self._query_id is None:
            return True

        try:
            result = self._client.stop_query(queryId=self._query_id)
            return result['success']
        except botocore.errorfactory.ClientError as err:
            already_stopped_msg = 'Query is not in Running or Scheduled state'
            if err.operation_name == 'StopQuery' \
                    and already_stopped_msg in err.__str__():
                return True
            raise err
        except Exception as err:
            raise err



class QueryNotYetStartError(Exception):
    """Raises error if the query has not yet started"""


class NotFetchQueryResultError(Exception):
    """Raises error if the query result could not be fetched"""


class ConfigInvalidSyntaxError(Exception):
    """Raises error if the configuration is invalid syntax"""


class ConfigVersionUnknownError(Exception):
    """Raises error if not specify configuration version"""


class ConfigNotFoundError(Exception):
    """Raises error if the configuration could not be found"""


class InvalidVersionError(Exception):
    """Raises error if the configuration version is invalid"""


class InvalidDurationError(Exception):
    """Raises error if the duration parameter is not number"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyinsights']

package_data = \
{'': ['*'], 'pyinsights': ['schema/*']}

install_requires = \
['boto3>=1.10.45,<2.0.0', 'jsonschema>=3.2.0,<4.0.0', 'pyyaml>=5.2,<6.0']

entry_points = \
{'console_scripts': ['pyinsights = pyinsights.cli:cli']}

setup_kwargs = {
    'name': 'pyinsights',
    'version': '0.1.3',
    'description': 'AWS CloudWatch Logs Insights is wrapped by Python',
    'long_description': "# PyInsights\n\nA CLI tool To query CloudWatch Logs Insights.\n\n## Usage\n\n### 1. Write Configuration\n\nwrite configuration to `pyinsights.yml` like:\n\n```yaml\nversion: '1.0'\nlog_group_name:\n  - '/ecs/sample'\nquery_string: 'field @message | filter @message like /ERROR/'\nduration: '30m'\nlimit: 10\n```\n\n### 2. Execute command\n\n```bash\npyinsights -c pyinsights.yml -p aws_profile -r region\n```\n",
    'author': 'homoluctus',
    'author_email': 'w.slife18sy@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/homoluctus/pyinsights',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys
import json
from time import sleep
import concurrent.futures as confu
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Dict

from pyinsights.aws import InsightsClient
from pyinsights.config import load_config, validate
from pyinsights.helper import get_times, processing


def query(params: Dict[str, Dict[str, Any]]) -> Dict[str, Any]:
    """Query to CloudWatch Logs Insights

    Arguments:
        params {Dict[str, Dict[str, Any]]}

    Returns:
        results {Dict[str, Any]}
    """

    client = InsightsClient(region=params['region'], profile=params['profile'])
    client.start_query(**params['config'])
    results = client.fetch_result()
    return results


def run_thread(params: Dict[str, Dict[str, Any]]) -> Dict[str, Any]:
    """Run thread

    Arguments:
        params {Dict[str, Dict[str, Any]]}

    Returns:
        results = Dict[str, Any]
    """

    with ThreadPoolExecutor(max_workers=1) as executor:
        thread = executor.submit(query, params)
        processing('Waiting', end=' ')

        while True:
            try:
                results = thread.result(timeout=0.1)
            except confu.TimeoutError:
                processing('.', end='')
                sleep(0.5)
                pass
            else:
                if results:
                    processing('.', end='\n')
                    break

    return results


def run(kwargs: Dict[str, str]) -> bool:
    """Run pyinsights

    Arguments:
        kwargs {Dict[str, Any]}

    Returns:
        bool
    """

    config = load_config(kwargs.pop('config'))
    validate(config)

    duration = config.pop('duration')
    if isinstance(duration, str):
        duration = get_times(duration)
    config.update(duration)
    kwargs.update({'config': config})

    results = run_thread(kwargs)
    json_results = json.dumps(results['results'], indent=2, ensure_ascii=False)
    sys.stdout.write(json_results)

    return True
